#include <iostream>
#include <fstream>
#include <memory>
#include <ctime>
#include <cstdarg>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <pwd.h>
#include "TomiTCP/net.h"
#include "TomiTCP/str.h"
#include "gate.h"
#include "log.h"
using namespace std;
using namespace net;

static unsigned short port = 6667, ts_port = 0, van_port = 0;
static string ts_host, van_host, setuid_user;

unsigned int reallen = 50;
unsigned int nicklen = 20;

auto_ptr<net::TomiTCP> ts_c, van_c;

bool debug = false;

static string config;
static int loadconfig(const char *fname);
static void rehash(int);
static int droppriv();

void ts_printf(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);

    vfprintf(*ts_c, format, ap);

    if (debug) {
	static bool nl = true;
	if (nl)
	    cout << "-->TS: ";
	nl = (format[strlen(format) - 1] == '\n');

	vprintf(format, ap);
    }
}

void van_printf(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);

    vfprintf(*van_c, format, ap);
    
    if (debug) {
	static bool nl = true;
	if (nl)
	    cout << "-->VAN: ";
	nl = (format[strlen(format) - 1] == '\n');

	vprintf(format, ap);
    }
}

void con_init()
{
    ts::con_init();
    vanilla::con_init();
}

void con_break()
{
    ts_printf("ERROR :(gate) Closing link: protocol problem\n");
    van_printf("ERROR :(gate) Closing link: protocol problem\n");
    ts_c->close();
    van_c->close();
}

int main(int argc, char *argv[])
{
    signal(SIGPIPE, SIG_IGN);
    signal(SIGHUP, rehash);
    initlog();

    /*
     * Parse command line arguments
     */
    {
        int arg = 1;
        while (arg < argc) {
            if (!strcmp(argv[arg],"-d"))
                debug = 1;
            else if (argv[arg][0] != '-' && !config.length())
                config = argv[arg];
            else {
                cerr << "Unrecognized option" << endl;
                return -1;
            }
            arg++;
        }
    }

    if (!config.length()) {
	cerr << "Usage: " << argv[0] << " [-d] <config.file>" << endl;
	return -1;
    }

    /*
     * Load config
     */
    if (loadconfig(config.c_str())) {
	cerr << "Config error" << endl;
	return -1;
    }

    /*
     * Drop privileges if requested
     */
    if (setuid_user.length()) {
	if (droppriv()) {
	    cerr << "Could not drop privileges" << endl;
	    return -1;
	}
    }

    /*
     * No, don't run this as root!
     */
    if (!getuid()) {
	cerr << "No, don't run this as root!" << endl;
	return -1;
    }

    /*
     * Go into background unless debug is set
     */
    if (!debug)
	daemon(1, 0);

    try {
	TomiTCP s(port);

	while (1) {
	    try {
		TomiTCP *n = s.accept();
		string host = tomi_ntop(n->rname), host2 = revers(n->rname);
		if (host == ts_host || host2 == ts_host) {
		    ts_c.reset(n);
		    try { van_c.reset(new TomiTCP(van_host, van_port)); }
		    catch (runtime_error e) {
			fprintf(*n, "ERROR :(gate) Closing link: Connect problem (%s)\n",
				e.what());
			throw;
		    }
		} else if (host == van_host || host2 == van_host) {
		    van_c.reset(n);
		    try { ts_c.reset(new TomiTCP(ts_host, ts_port)); }
		    catch (runtime_error e) {
			fprintf(*n, "ERROR :(gate) Closing link: Connect problem (%s)\n",
				e.what());
			throw;
		    }
		} else {
		    fprintf(*n, "ERROR :(gate) Closing link: %s (Unathorized connection)\n",
			    host.c_str());
		    delete n;
		    throw runtime_error("unauthorized connection - " + host);
		}

		con_init();

		if (debug)
		    cout << "NEW" << endl;

		string buf;
		fd_set set;
		while (ts_c->ok() && van_c->ok()) {
		    FD_ZERO(&set);
		    FD_SET(ts_c->sock, &set);
		    FD_SET(van_c->sock, &set);

		    int ret = TEMP_FAILURE_RETRY(select(FD_SETSIZE, &set, 0, 0, 0));
		    if (ret == -1)
			throw runtime_error("error in select: " + string(strerror(errno)));

		    /*
		     * Well, I know that doing a getline here is not really a
		     * good idea, but it shouldn't be a big problem. It's not
		     * a good idea to bother incompatible ircd's with
		     * nickchange and mode races anyway, so you shouldn't have
		     * a lossy and laggy line between them. So, as we expect
		     * you have, we can just getline ;). -lis-
		     */
		    if (FD_ISSET(ts_c->sock, &set)) {
			if (!ts_c->getline(buf))
			    throw runtime_error("ts - getline error (maybe EOF)");
			chomp(buf);

			if (debug)
			    cout << "<--TS: " << buf << endl;

			string prefix;
			vector<string> cmd;
			parsein(buf, prefix, cmd);

			if (cmd.size()) {
			    strtoupper(cmd[0]);
			    ts::line(prefix, cmd);
			}
		    }

		    if (FD_ISSET(van_c->sock, &set)) {
			if (!van_c->getline(buf))
			    throw runtime_error("vanilla - getline error (maybe EOF)");
			chomp(buf);

			if (debug)
			    cout << "<--VAN: " << buf << endl;

			string prefix;
			vector<string> cmd;
			parsein(buf, prefix, cmd);

			if (cmd.size()) {
			    strtoupper(cmd[0]);
			    vanilla::line(prefix, cmd);
			}
		    }
		}
	    } catch (runtime_error e) {
		log(LOG_WARNING, "Error: %s", e.what());
	    }

	    ts_c.reset(0);
	    van_c.reset(0);
	}
    } catch (runtime_error e) {
	log(LOG_WARNING, "Error: %s", e.what());
    }

    return 0;
}

static int loadconfig(const char *fname)
{
    ifstream in(fname);
    if (!in) {
	log(LOG_WARNING, "Cannot open config file %s", fname);
	return -1;
    }

    string l; int ln = 0;
    while (getline(in, l)) {
	ln++;

	if (l.length() && l[0] != '#') {
	    unsigned int d = l.find_first_of(" \f\n\r\t\v");
	    if (d > 0 && d != string::npos) {
		string a(l,0,d), b(l,d+1);
		wstrip(a); wstrip(b);
		strtolower(a);

		if (a == "van_host")
		    van_host = b;
		else if (a == "ts_host")
		    ts_host = b;
		else if (a == "setuid_user")
		    setuid_user = b;
		else if (a == "van_port")
		    van_port = atol(b.c_str());
		else if (a == "ts_port")
		    ts_port = atol(b.c_str());
		else if (a == "port")
		    port = atol(b.c_str());
		else if (a == "reallen")
		    reallen = atol(b.c_str());
		else if (a == "nicklen")
		    nicklen = atol(b.c_str());
		else
		    log(LOG_WARNING, "Error on line %i (%s is unknown option)",
			    ln, a.c_str());
	    } else {
		log(LOG_WARNING, "Error on line %i", ln);
	    }
	}
    }

    return 0;
}

static void rehash(int)
{
    log(LOG_NOTICE, "rehashing...");
    loadconfig(config.c_str());
}

static int droppriv()
{
    struct passwd *p = getpwnam(setuid_user.c_str());
    if (!p) {
	log(LOG_WARNING, "Specified user does not exist");
	return -1;
    }

    if (setgid(p->pw_gid)) {
	log(LOG_WARNING, "Could not setgid");
	return -1;
    }

    if (setuid(p->pw_uid)) {
	log(LOG_WARNING, "Could not setuid");
	return -1;
    }

    return 0;
}
